
import json
import os
from typing import Any, Dict, List

import pytest
from attr import define

from featureflags.evaluations.auth_target import Target
from featureflags.evaluations.evaluator import Evaluator
from featureflags.evaluations.feature import FeatureConfig, FeatureConfigKind
from featureflags.evaluations.segment import Segment
from featureflags.lru_cache import LRUCache
from featureflags.repository import Repository

BASE_PATH = os.path.dirname(os.path.realpath(__file__))
PATH = cwd = BASE_PATH + "/ff-test-cases/tests/"


cache = LRUCache()
repository = Repository(cache)
evaluator = Evaluator(repository)


@define
class Usecase:
    flag: FeatureConfig
    segments: List[Segment]
    targets: List[Target]
    expected: Dict[str, Any]

    @classmethod
    def from_dict(cls, source: Dict[str, Any]) -> 'Usecase':
        flag = FeatureConfig.from_dict(source["flag"])

        segments = []
        if "segments" in source:
            segments = [Segment.from_dict(segment)
                        for segment in source["segments"]]
        targets = [Target.from_dict(target) for target in source["targets"]]
        expected = source["expected"]
        return Usecase(flag=flag, segments=segments, targets=targets,
                       expected=expected)


def load_json_test_file(filename: str) -> Usecase:
    result = None
    with open(filename, 'r') as f:
        result = Usecase.from_dict(json.load(f))
    return result


def load_test_files():
    result = []
    for file in os.listdir(PATH):
        if file.endswith(".json"):
            fname = os.path.join(PATH, file)
            usecase = load_json_test_file(fname)
            usecase.flag.feature += file
            result.append((fname, usecase))
    return result


def usecase_provider():
    usecases = load_test_files()

    result = []
    for fname, usecase in usecases:
        repository.set_flag(usecase.flag)

        for segment in usecase.segments:
            repository.set_segment(segment)

        for identifier, value in usecase.expected.items():
            result.append((fname, identifier, value, usecase))

    return result


@pytest.mark.parametrize('fname,identifier,expected,usecase',
                         usecase_provider())
def test_evaluator(fname, identifier, expected, usecase):
    target = None
    if identifier != "_no_target":
        target = next(
            (val for val in usecase.targets if val.identifier == identifier),
            None
        )

    got = None

    switch_kind = {
        FeatureConfigKind.BOOLEAN:
            lambda: evaluator.evaluate(usecase.flag.feature, target).bool(),
        FeatureConfigKind.STRING:
            lambda: evaluator.evaluate(usecase.flag.feature, target).string(),
        FeatureConfigKind.INT:
            lambda: evaluator.evaluate(usecase.flag.feature, target).int(),
        FeatureConfigKind.NUMBER:
            lambda: evaluator.evaluate(usecase.flag.feature, target).number(),
        FeatureConfigKind.JSON:
            lambda: evaluator.evaluate(usecase.flag.feature, target).json(),
    }

    got = switch_kind[usecase.flag.kind]()

    assert got == expected



import time
from threading import Event, Thread

from featureflags.api.client import AuthenticatedClient
from featureflags.repository import DataProviderInterface

from .api.default.get_all_segments import sync as retrieve_segments
from .api.default.get_feature_config import sync as retrieve_flags
from .config import Config
from .util import log


class PollingProcessor(Thread):

    def __init__(self, client: AuthenticatedClient, config: Config,
                 environment_id: str, ready: Event,
                 stream_ready: Event,
                 repository: DataProviderInterface) -> None:
        Thread.__init__(self)
        self.daemon = True
        self.__environment_id = environment_id
        self.__client = client
        self.__config = config
        self.__running = False
        self.__ready = ready
        self.__stream_ready = stream_ready
        self.__repository = repository

    def run(self):
        if not self.__running:
            if self.__config.pull_interval < 60:
                log.warning("Pull Interval must be greater than or equal "
                            "to 60 seconds, was: " +
                            str(self.__config.pull_interval) +
                            " setting to 60")
                self.__config.pull_interval = 60
            log.info("Starting PollingProcessor with request interval: " +
                     str(self.__config.pull_interval))
            self.__running = True
            while self.__running:
                start_time = time.time()
                try:
                    t1 = Thread(target=self.__retrieve_segments)
                    t2 = Thread(target=self.__retrieve_flags)
                    t1.start()
                    t2.start()
                    t1.join()
                    t2.join()
                    if not self.__ready.is_set() is True:
                        log.info("PollingProcessor initialized ok")
                        if self.__config.enable_stream and \
                                not self.__stream_ready.is_set():
                            log.debug('Poller is in pause mode...')
                            self.__ready.wait()
                        else:
                            self.__ready.set()
                except Exception as e:
                    log.exception(
                        'Error: Exception encountered when polling flags. %s',
                        e
                    )

                elapsed = time.time() - start_time
                if elapsed < self.__config.pull_interval:
                    time.sleep(self.__config.pull_interval - elapsed)

    def stop(self):
        log.info("Stopping PollingProcessor")
        self.__running = False

    def __retrieve_flags(self):
        log.debug("Loading feature flags")
        flags = retrieve_flags(
            client=self.__client, environment_uuid=self.__environment_id
        )
        log.debug("Feature flags loaded")
        for flag in flags:
            log.debug("Put flag %s into repository", flag.feature)
            self.__repository.set_flag(flag)

    def __retrieve_segments(self):
        log.debug("Loading target segments")
        segments = retrieve_segments(
            client=self.__client, environment_uuid=self.__environment_id
        )
        log.debug("Target segments loaded")
        for segment in segments:
            log.debug("Put %s segment into repository", segment.identifier)
            self.__repository.set_segment(segment)



import pytest

from featureflags.evaluations.clause import Clause
from featureflags.evaluations.constants import (CONTAINS_OPERATOR,
                                                ENDS_WITH_OPERATOR,
                                                EQUAL_OPERATOR,
                                                EQUAL_SENSITIVE_OPERATOR,
                                                GT_OPERATOR, IN_OPERATOR,
                                                STARTS_WITH_OPERATOR)
from featureflags.ftypes import String


@pytest.mark.parametrize(
    "op",
    [
        None,
        "NOT_FOUND",
    ],
)
def test_evaluate_op(op):
    clause = Clause(
        attribute="email", id="", negate=False, op=op, values=["john@doe.com"]
    )

    got = clause.evaluate(None, None, String("john@doe.com"))

    assert got is False


@pytest.mark.parametrize(
    "op,method,expected",
    [
        (IN_OPERATOR, "in_list", True),
        (EQUAL_OPERATOR, "equal", True),
        (GT_OPERATOR, "greater_than", False),
        (STARTS_WITH_OPERATOR, "starts_with", True),
        (ENDS_WITH_OPERATOR, "ends_with", True),
        (CONTAINS_OPERATOR, "contains", True),
        (EQUAL_SENSITIVE_OPERATOR, "equal_sensitive", True),
    ],
)
def test_evaluate_string(mocker, op, method, expected):
    clause = Clause(
        attribute="email", id="", negate=False, op=op, values=["john@doe.com"]
    )

    m = mocker.patch.object(String, method, return_value=expected)
    _string = String("john@doe.com")

    got = clause.evaluate(None, None, _string)

    assert got == expected
    assert m.call_count == 1



import re
import typing

import attr

from featureflags.ftypes.interface import Interface

from .utils import get_str_value


@attr.s(auto_attribs=True)
class String(Interface):

    value: str

    def starts_with(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return self.value.startswith(_value)

    def ends_with(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return self.value.endswith(_value)

    def match(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return re.match(self.value, _value) is not None

    def contains(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return _value in self.value

    def equal_sensitive(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return self.value == _value

    def equal(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return self.value.lower() == _value.lower()

    def greater_than(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return self.value.lower() > _value.lower()

    def greater_than_equal(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return self.value.lower() >= _value.lower()

    def less_than(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return self.value.lower() < _value.lower()

    def less_than_equal(self, value: typing.Any) -> bool:
        _value = get_str_value(value)
        return self.value.lower() <= _value.lower()

    def in_list(self, value: typing.List[typing.Any]) -> bool:
        return self.value.lower() in (val.lower() for val in value)



from typing import List, Optional, Union

import mmh3

from featureflags.evaluations.auth_target import Target
from featureflags.evaluations.clause import Clause, Clauses
from featureflags.evaluations.constants import (CONTAINS_OPERATOR,
                                                ENDS_WITH_OPERATOR,
                                                EQUAL_OPERATOR,
                                                EQUAL_SENSITIVE_OPERATOR,
                                                GT_OPERATOR, IN_OPERATOR,
                                                ONE_HUNDRED,
                                                SEGMENT_MATCH_OPERATOR,
                                                STARTS_WITH_OPERATOR)
from featureflags.evaluations.distribution import Distribution
from featureflags.evaluations.enum import FeatureState
from featureflags.evaluations.feature import FeatureConfig
from featureflags.evaluations.serving_rule import ServingRule, ServingRules
from featureflags.evaluations.variation import Variation
from featureflags.evaluations.variation_map import VariationMap
from featureflags.models.unset import Unset
from featureflags.repository import QueryInterface
from featureflags.util import log

EMPTY_VARIATION = Variation(identifier="", value=None)


class Evaluator(object):

    def __init__(self, provider: QueryInterface):
        self.provider = provider

    def _find_variation(self, variations: List[Variation],
                        identifier: Optional[str]) -> Variation:
        if not identifier:
            log.debug("Empty identifier %s or variations %s occurred",
                      identifier, variations)
            return EMPTY_VARIATION
        variation = next(
            (val for val in variations if val.identifier == identifier),
            EMPTY_VARIATION
        )
        log.debug("Variation %s found in variations %s",
                  identifier, variations)
        return variation

    def _get_normalized_number(self, bucket_by: str, identifier: str):
        value = ":".join([bucket_by, identifier])
        hash = int(mmh3.hash(value))
        normalized_number = (hash % ONE_HUNDRED) + 1
        log.debug("normalized number for %s = %d", value, normalized_number)
        return normalized_number

    def _is_enabled(self, target: Target, bucket_by: str,
                    percentage: int) -> bool:
        attr_value = target.get_attr_value(bucket_by)
        if not attr_value:
            log.debug("Returns False. %s is set to %s", bucket_by, attr_value)
            return False
        bucket_id = self._get_normalized_number(bucket_by, attr_value)
        log.debug("Bucket id %d for target %s", bucket_id, target.identifier)
        return percentage > 0 and bucket_id <= percentage

    def _evaluate_distribution(self, distribution: Distribution,
                               target: Target) -> Optional[str]:
        variation = None
        if not distribution:
            log.debug("Distribution is empty")
            return variation
        total_percentage = 0
        for _variation in distribution.variations:
            variation = _variation.variation
            log.debug("Checking variation %s", variation)
            total_percentage += _variation.weight
            if self._is_enabled(target, distribution.bucket_by,
                                total_percentage):
                log.debug("Enabled for distribution %s", distribution)
                return variation
        log.debug("Variation of distribution evaluation %s", variation)
        return variation

    def _check_target_in_segment(self, segments: List[str],
                                 target: Target) -> bool:
        for segment_identifier in segments:
            segment = self.provider.get_segment(segment_identifier)

            if segment:
                # Should Target be excluded - if in excluded
                # list we return false
                if not isinstance(segment.excluded, Unset) and \
                        next(
                            (val for val in segment.excluded
                             if val.identifier == target.identifier),
                            None) is not None:
                    log.debug('Target % s excluded from segment % s' +
                              'via exclude list\n',
                              target.name, segment.name)
                    return False

                # Should Target be included - if in included list
                #  we return true
                log.info("target %s, segment %s",
                         target.identifier, segment.included)
                if not isinstance(segment.included, Unset) and \
                        next(
                            (val for val in segment.included
                             if val.identifier == target.identifier),
                            None) is not None:
                    log.debug('Target %s included in segment %s' +
                              ' via include list\n',
                              target.name, segment.name)
                    return True

                # Should Target be included via segment rules
                if segment.rules and self._evaluate_clauses(segment.rules,
                                                            target):
                    log.debug('Target %s included in segment %s via rules\n',
                              target.name, segment.name)
                    return True
        log.debug("Target groups empty return false")
        return False

    def _evaluate_clause(self, clause: Clause, target: Target) -> bool:
        if clause.op is None:
            log.debug("Clause is empty")
            return False

        if not clause.values:
            log.debug("Clause values is empty")
            return False

        operator = clause.op.lower()
        type = target.get_type(clause.attribute)

        if type is None:
            if operator == SEGMENT_MATCH_OPERATOR.lower():
                log.debug("Clause operator is %s, evaluate on segment",
                          operator)
                return self._check_target_in_segment(clause.values, target)
            log.debug("Attribute type %s is none return false", type)
            return False
        log.debug("evaluate clause with object %s operator %s and value %s",
                  object, operator, clause.values)
        if operator == IN_OPERATOR.lower():
            return type.in_list(clause.values)
        if operator == EQUAL_OPERATOR.lower():
            return type.equal(clause.values)
        if operator == GT_OPERATOR.lower():
            return type.greater_than(clause.values)
        if operator == STARTS_WITH_OPERATOR.lower():
            return type.starts_with(clause.values)
        if operator == ENDS_WITH_OPERATOR.lower():
            return type.ends_with(clause.values)
        if operator == CONTAINS_OPERATOR.lower():
            return type.contains(clause.values)
        if operator == EQUAL_SENSITIVE_OPERATOR.lower():
            return type.equal_sensitive(clause.values)
        # unknown operation
        return False

    def _evaluate_clauses(self, clauses: Union[Unset, Clauses],
                          target: Target) -> bool:
        if not isinstance(clauses, Unset):
            for clause in clauses:
                if not self._evaluate_clause(clause, target):
                    log.debug("Unsuccessful evaluation of clause %s", clause)
                    return False
        log.debug("All clauses %s evaluated", clauses)
        return True

    def _evaluate_rule(self, rule: ServingRule, target: Target) -> bool:
        return self._evaluate_clauses(rule.clauses, target)

    def _evaluate_rules(self, rules: ServingRules,
                        target: Target) -> Optional[str]:
        if not rules or not target:
            log.debug("There is no target or serving rule, %s, %s",
                      rules, target)
            return None

        def sort_by_priority(item: ServingRule):
            return item.priority

        log.debug("Sorting serving rules %s", rules)
        rules.sort(key=sort_by_priority)
        log.debug("Sorted serving rules %s", rules)

        identifier = None
        for rule in rules:
            # if evaluation is false just continue to next rule
            if not self._evaluate_rule(rule, target):
                log.debug(
                    "Unsuccessful evaluation of rule %s continue to next rule",
                    rule)
                continue

            # rule matched, check if there is distribution
            distribution = rule.serve.distribution
            if not isinstance(distribution, Unset) and distribution:
                log.debug("Evaluate distribution %s", distribution)
                identifier = self._evaluate_distribution(
                    distribution, target)

            # rule matched, here must be variation if distribution is None
            variation = rule.serve.variation
            if not isinstance(variation, Unset) and variation:
                log.debug("Return rule variation identifier %s", identifier)
                identifier = variation

            return identifier

        log.debug("All rules failed, return empty variation identifier")
        return None

    def _evaluate_variation_map(self, var_target_map: List[VariationMap],
                                target: Target) -> Optional[str]:
        if not target or not var_target_map:
            log.debug("Target is none")
            return None

        for variation_map in var_target_map:
            if not isinstance(variation_map.targets, Unset) and next(
                (val for val in variation_map.targets
                    if not isinstance(val, Unset) and val.identifier ==
                 target.identifier), None) is not None:
                log.debug("Evaluate variation map with result %s",
                          variation_map.variation)
                return variation_map.variation

            segment_identifiers = variation_map.target_segments
            if not isinstance(segment_identifiers, Unset) and \
                    self._check_target_in_segment(segment_identifiers, target):
                log.debug("Evaluate variationMap with segment " +
                          "identifiers % s and return % s",
                          segment_identifiers, variation_map.variation)
                return variation_map.variation

        return None

    def _evaluate_flag(self, fc: FeatureConfig,
                       target: Target) -> Variation:
        variation: Optional[str] = fc.off_variation
        log.debug("feature %s state is %s", fc.feature, fc.state)
        if fc.state == FeatureState.ON:
            variation = None
            if not isinstance(fc.variation_to_target_map, Unset):
                variation = self._evaluate_variation_map(
                    fc.variation_to_target_map, target)
                log.debug("variation %s found in target map", variation)

            if not variation:
                variation = self._evaluate_rules(fc.rules, target)
                log.debug("variation %s found in rules", variation)

            if not variation and not isinstance(fc.default_serve.distribution,
                                                Unset):
                variation = self._evaluate_distribution(
                    fc.default_serve.distribution, target)
                log.debug(
                    "variation %s found in default serve distribution",
                    variation)

            if not variation and not isinstance(fc.default_serve.variation,
                                                Unset):
                variation = fc.default_serve.variation
                log.debug("variation %s found in default serve", variation)

        return self._find_variation(fc.variations, variation)

    def _check_prerequisite(self, parent: FeatureConfig,
                            target: Target) -> bool:
        if not isinstance(parent.prerequisites, Unset):
            log.info('Checking pre requisites %s of parent feature %s',
                     parent.prerequisites, parent.feature)
            for pqs in parent.prerequisites:
                config = self.provider.get_flag(pqs.feature)
                if not config:
                    log.warning(
                        'Could not retrieve the pre requisite details of ' +
                        'feature flag: %s',
                        parent.feature)
                    return True

                # Pre requisite variation value evaluated below
                variation = self._evaluate_flag(config, target)
                log.info('Pre requisite flag %s has variation %s ' +
                         'for target %s',
                         config.feature, variation, target)

                # Compare if the pre requisite variation is a possible
                # valid value of the pre requisite FF
                log.info('Pre requisite flag %s should have the variations %s',
                         config.feature, pqs.variations)

                if not isinstance(variation, Unset) and variation.identifier \
                        not in pqs.variations:
                    return False
                else:
                    return self._check_prerequisite(config, target)
        return True

    def evaluate(self, identifier: str, target: Target) -> Variation:
        fc = self.provider.get_flag(identifier)
        if not fc:
            return Variation(identifier="", value=None)

        if fc.prerequisites:
            prereq = self._check_prerequisite(fc, target)
            if not prereq:
                return self._find_variation(fc.variations, fc.off_variation)

        return self._evaluate_flag(fc, target)



from typing import Dict

import attr


@attr.s(auto_attribs=True)
class Client:
    """A class for keeping track of data related to the API"""

    base_url: str
    events_url: str
    params: Dict[str, str] = attr.ib(factory=dict, kw_only=True)
    cookies: Dict[str, str] = attr.ib(factory=dict, kw_only=True)
    headers: Dict[str, str] = attr.ib(factory=dict, kw_only=True)
    timeout: float = attr.ib(5.0, kw_only=True)

    def get_headers(self) -> Dict[str, str]:
        """Get headers to be used in all endpoints"""
        return {**self.headers}

    def with_headers(self, headers: Dict[str, str]) -> "Client":
        """Get a new client matching this one with additional headers"""
        return attr.evolve(self, headers={**self.headers, **headers})

    def get_params(self) -> Dict[str, str]:
        return {**self.params}

    def get_cookies(self) -> Dict[str, str]:
        return {**self.cookies}

    def with_cookies(self, cookies: Dict[str, str]) -> "Client":
        """Get a new client matching this one with additional cookies"""
        return attr.evolve(self, cookies={**self.cookies, **cookies})

    def get_timeout(self) -> float:
        return self.timeout

    def with_timeout(self, timeout: float) -> "Client":
        """
        Get a new client matching this one with a new timeout (in seconds)
        """
        return attr.evolve(self, timeout=timeout)


@attr.s(auto_attribs=True)
class AuthenticatedClient(Client):
    """A Client which has been authenticated for use on secured endpoints"""

    token: str

    def get_headers(self) -> Dict[str, str]:
        """Get headers to be used in authenticated endpoints"""
        return {"Authorization": f"Bearer {self.token}", **self.headers}



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from featureflags.models import UNSET, Unset

from .distribution import Distribution

T = TypeVar("T", bound="Serve")


@attr.s(auto_attribs=True)
class Serve(object):
    distribution: Union[Unset, Distribution] = UNSET
    variation: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        distribution: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.distribution, Unset):
            distribution = self.distribution.to_dict()

        variation = self.variation

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if distribution is not UNSET:
            field_dict["distribution"] = distribution
        if variation is not UNSET:
            field_dict["variation"] = variation

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        _distribution = d.pop("distribution", UNSET)
        distribution: Union[Unset, Distribution]
        if isinstance(_distribution, Unset):
            distribution = UNSET
        else:
            distribution = Distribution.from_dict(_distribution)

        variation = d.pop("variation", UNSET)

        serve = cls(
            distribution=distribution,
            variation=variation,
        )

        serve.additional_properties = d
        return serve

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
from typing import Any, Dict, Type, TypeVar

import attr

T = TypeVar("T", bound="Message")


@attr.s(auto_attribs=True)
class Message(object):
    event: str
    domain: str
    identifier: str
    version: int

    def to_dict(self) -> Dict[str, Any]:
        event = self.event
        domain = self.domain
        identifier = self.identifier
        version = self.version

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "event": event,
                "domain": domain,
                "identifier": identifier,
                "version": version
            }
        )
        return field_dict

    @classmethod
    def from_str(cls: Type[T], src: str) -> T:
        if src is None or src == '':
            raise ValueError('source cannot be empty or None')
        d = json.loads(src)
        return cls.from_dict(d)

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        event = d.pop("event")
        domain = d.pop("domain")
        identifier = d.pop("identifier")
        version = d.pop("version")

        message = cls(
            event=event,
            domain=domain,
            identifier=identifier,
            version=version
        )

        return message



import abc
from typing import List, Optional

from featureflags.evaluations.constants import SEGMENT_MATCH_OPERATOR
from featureflags.evaluations.feature import FeatureConfig
from featureflags.evaluations.segment import Segment
from featureflags.interface import Cache, Store
from featureflags.models.unset import Unset
from featureflags.util import log


class QueryInterface(metaclass=abc.ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_flag') and
                callable(subclass.get_flag) and
                hasattr(subclass, 'get_segment') and
                callable(subclass.get_segment) or
                hasattr(subclass, 'find_flags_by_segment') and
                callable(subclass.find_flags_by_segment) or
                NotImplemented)

    @abc.abstractmethod
    def get_flag(self, identifier: str) -> Optional[FeatureConfig]:
        """Get flag from repository"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_segment(self, identifier: str) -> Optional[Segment]:
        """Get Target group from repository"""
        raise NotImplementedError

    @abc.abstractmethod
    def find_flags_by_segment(self, identifier: str) -> List[str]:
        """Find all flags with rule segment match"""
        raise NotImplementedError


class DataProviderInterface(QueryInterface):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_flag') and
                callable(subclass.get_flag) and
                hasattr(subclass, 'get_segment') and
                callable(subclass.get_segment) or
                hasattr(subclass, 'close') and
                callable(subclass.close) or
                NotImplemented)

    @abc.abstractmethod
    def set_flag(self, flag: FeatureConfig) -> None:
        """Put flag to the repository"""
        raise NotImplementedError

    @abc.abstractmethod
    def set_segment(self, group: Segment) -> None:
        """Put Target group to the repository"""
        raise NotImplementedError

    @abc.abstractmethod
    def remove_flag(self, identifier: str) -> None:
        """Remove Flag from the repository"""
        raise NotImplementedError

    @abc.abstractmethod
    def remove_segment(self, identifier: str) -> None:
        """Remove Target group from the repository"""
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """Put Target group to the repository"""
        raise NotImplementedError


class Repository(DataProviderInterface):

    def __init__(self, cache: Cache, store: Store = None) -> None:
        self.cache = cache
        self.store = store

    def get_flag(self, identifier: str,
                 cacheable: bool = None) -> Optional[FeatureConfig]:
        flag_key = format_flag_key(identifier)
        try:
            flag = self.cache.get(flag_key)
            log.debug("return flag from cache %s", identifier)
            return flag
        except KeyError:
            if self.store:
                flag = self.store.get(flag_key)
                log.debug("return flag from store %s", identifier)
                if flag and cacheable:
                    log.debug("set flag to the cache %s", identifier)
                    self.cache.set(flag_key, flag)
                return flag
        log.warning("flag not found %s", identifier)
        return None

    def get_segment(self, identifier: str,
                    cacheable: bool = None) -> Optional[Segment]:
        segment_key = format_segment_key(identifier)
        try:
            segment = self.cache.get(segment_key)
            log.debug("return segment from cache %s", identifier)
            return segment
        except KeyError:
            if self.store:
                segment = self.store.get(segment_key)
                log.debug("return segment from store %s", identifier)
                if segment and cacheable:
                    log.debug("set segment to the cache %s", identifier)
                    self.cache.set(segment_key, segment)
                return segment
        log.warning("segment not found %s", identifier)
        return None

    def set_flag(self, flag: FeatureConfig) -> None:
        if self.is_flag_outdated(flag.feature, flag):
            log.debug("Flag %s already exists", flag.feature)
            return None

        flag_key = format_flag_key(flag.feature)

        if self.store:
            self.store.set(flag_key, flag)
            self.cache.remove([flag_key])
            log.debug(
                "Flag %s successfully stored and cache invalidated",
                flag.feature)
        else:
            self.cache.set(flag_key, flag)
            log.debug("Flag %s successfully cached", flag.feature)

    def set_segment(self, segment: Segment) -> None:
        if self.is_segment_outdated(segment.identifier, segment):
            log.debug("Segment %s already exists", segment.identifier)
            return None

        segment_key = format_segment_key(segment.identifier)

        if self.store:
            self.store.set(segment_key, segment)
            self.cache.remove([segment_key])
            log.debug(
                "Segment %s successfully stored and cache invalidated",
                segment.identifier)
        else:
            self.cache.set(segment_key, segment)
            log.debug("Segment %s successfully cached", segment.identifier)

    def find_flags_by_segment(self, segment: str) -> List[str]:
        result = []
        keys = self.cache.keys()
        if self.store:
            keys = self.store.keys()
        for key in keys:
            flag = self.get_flag(key, cacheable=False)
            if not flag:
                continue
            for serving_rule in flag.rules:
                for clause in serving_rule.clauses:
                    if clause.op == SEGMENT_MATCH_OPERATOR and not next(
                        (val for val in clause.values if val == segment),
                        None
                    ):
                        log.debug("Flag %s evaluated in segments",
                                  flag.feature)
                        result.append(flag.feature)
        return result

    def remove_flag(self, identifier: str) -> None:
        """Remove Flag from the repository"""
        flag_key = format_flag_key(identifier)
        if self.store:
            self.store.remove([flag_key])
            log.debug("Flag %s successfully deleted from store", identifier)

        self.cache.remove([flag_key])
        log.debug("Flag %s successfully deleted from cache", identifier)

    def remove_segment(self, identifier: str) -> None:
        """Remove Target group from the repository"""
        segment_key = format_segment_key(identifier)
        if self.store:
            self.store.remove([segment_key])
            log.debug("Segment %s successfully deleted from store", identifier)

        self.cache.remove([segment_key])
        log.debug("Segment %s successfully deleted from cache", identifier)

    def close(self) -> None:
        if self.store:
            self.store.close()

    def is_flag_outdated(self, identifier: str,
                         new_config: FeatureConfig) -> bool:
        flag = self.get_flag(identifier, cacheable=False)
        if flag and not isinstance(flag.version, Unset) and \
                not isinstance(new_config, Unset) and \
                not isinstance(new_config.version, Unset):
            return flag.version >= new_config.version
        return False

    def is_segment_outdated(self, identifier: str,
                            new_segment: Segment) -> bool:
        segment = self.get_segment(identifier, cacheable=False)
        if segment and not isinstance(segment.version, Unset) and \
                not isinstance(new_segment, Unset) and \
                not isinstance(new_segment.version, Unset):
            return segment.version >= new_segment.version
        return False


def format_flag_key(identifier: str) -> str:
    return f'flags/{identifier}'


def format_segment_key(identifier: str) -> str:
    return f'segments/{identifier}'



import typing

import attr

from .interface import Interface
from .utils import get_value


@attr.s(auto_attribs=True)
class Boolean(Interface):

    value: bool

    def starts_with(self, value: typing.Any) -> bool:
        return False

    def ends_with(self, value: typing.Any) -> bool:
        return False

    def match(self, value: typing.Any) -> bool:
        return False

    def contains(self, value: typing.Any) -> bool:
        return False

    def equal_sensitive(self, value: typing.Any) -> bool:
        return False

    def equal(self, value: typing.Any) -> bool:
        _value = get_value(value)
        return self.value is _value

    def greater_than(self, value: typing.Any) -> bool:
        return False

    def greater_than_equal(self, value: typing.Any) -> bool:
        return False

    def less_than(self, value: typing.Any) -> bool:
        return False

    def less_than_equal(self, value: typing.Any) -> bool:
        return False

    def in_list(self, value: typing.List[typing.Any]) -> bool:
        return False



from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import attr

from featureflags.ftypes import TYPES
from featureflags.ftypes.interface import Interface
from featureflags.models import UNSET, Unset
from featureflags.util import log

from .target_attributes import TargetAttributes

T = TypeVar("T", bound="Target")


@attr.s(auto_attribs=True)
class Target(object):

    identifier: str
    account: str
    org: str
    environment: str
    project: str
    name: str
    anonymous: Union[Unset, bool] = UNSET
    attributes: Union[Unset, TargetAttributes] = UNSET
    created_at: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier
        account = self.account
        org = self.org
        environment = self.environment
        project = self.project
        name = self.name
        anonymous = self.anonymous
        attributes: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.attributes, Unset):
            attributes = self.attributes.to_dict()

        created_at = self.created_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
                "account": account,
                "org": org,
                "environment": environment,
                "project": project,
                "name": name,
            }
        )
        if anonymous is not UNSET:
            field_dict["anonymous"] = anonymous
        if attributes is not UNSET:
            field_dict["attributes"] = attributes
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        identifier = d.pop("identifier")

        account = d.pop("account")

        org = d.pop("org")

        environment = d.pop("environment")

        project = d.pop("project")

        name = d.pop("name")

        anonymous = d.pop("anonymous", UNSET)

        _attributes = d.pop("attributes", UNSET)
        attributes: Union[Unset, TargetAttributes]
        if isinstance(_attributes, Unset):
            attributes = UNSET
        else:
            attributes = TargetAttributes.from_dict(_attributes)

        created_at = d.pop("createdAt", UNSET)

        target = cls(
            identifier=identifier,
            account=account,
            org=org,
            environment=environment,
            project=project,
            name=name,
            anonymous=anonymous,
            attributes=attributes,
            created_at=created_at,
        )

        target.additional_properties = d
        return target

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties

    def get_attr_value(self, attribute: str) -> Optional[str]:
        if not attribute:
            log.debug("Attribute is empty")
            return None
        result: Any = getattr(self, attribute, None)
        if isinstance(result, Unset) and not isinstance(self.attributes,
                                                        Unset):
            log.debug("Checking attributes field %s", attribute)
            result = self.attributes.get(attribute, None)
            if not result:
                log.warning("Attribute %s does not exist", attribute)

        return result

    def get_type(self, attribute: str) -> Optional[Interface]:
        value: Optional[str] = self.get_attr_value(attribute)
        for _type, klass in TYPES.items():
            if isinstance(value, _type):
                operator = TYPES.get(_type, None)
                if operator:
                    return klass(value)
        log.debug("Unsupported type found on attribute %s", attribute)
        return None



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from .authentication_request_target import AuthenticationRequestTarget
from .unset import UNSET, Unset

T = TypeVar("T", bound="AuthenticationRequest")


@attr.s(auto_attribs=True)
class AuthenticationRequest:
    """ """

    api_key: str
    target: Union[Unset, AuthenticationRequestTarget] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        api_key = self.api_key
        target: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.target, Unset):
            target = self.target.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "apiKey": api_key,
            }
        )
        if target is not UNSET:
            field_dict["target"] = target

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        api_key = d.pop("apiKey")

        _target = d.pop("target", UNSET)
        target: Union[Unset, AuthenticationRequestTarget]
        if isinstance(_target, Unset):
            target = UNSET
        else:
            target = AuthenticationRequestTarget.from_dict(_target)

        authentication_request = cls(
            api_key=api_key,
            target=target,
        )

        authentication_request.additional_properties = d
        return authentication_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections import OrderedDict
from typing import Any, List

from .interface import Cache


class LRUCache(Cache):
    def __init__(self, *args: Any, size: int = 1000, **kwargs: Any) -> None:
        self.size = size
        init = args
        if len(init) > 0:
            init = args[0][-size:]
        self.cache: OrderedDict = OrderedDict(init)

    def __contains__(self, key: str) -> bool:
        return self.get(key) is not None

    def __getitem__(self, key: str) -> Any:
        return self.get(key)

    def __setitem__(self, key: str, value: Any) -> Any:
        self.cache[key] = value
        self.cache.move_to_end(key)

        while len(self.cache) > self.size:
            oldkey = next(iter(self.cache))
            del self.cache[oldkey]

    def __len__(self) -> int:
        return len(self.cache)

    def set(self, key: str, value: Any) -> None:
        self.__setitem__(key, value)

    def get(self, key: str) -> Any:
        val = self.cache.get(key)
        self.cache.move_to_end(key)
        return val

    def remove(self, keys: List[str]) -> None:
        for key in keys:
            del self.cache[key]

    def keys(self) -> List[str]:
        return list(self.cache.keys())



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.key_value import KeyValue
from ..models.metrics_data_metrics_type import MetricsDataMetricsType

T = TypeVar("T", bound="MetricsData")


@attr.s(auto_attribs=True)
class MetricsData:
    """ """

    timestamp: int
    count: int
    metrics_type: MetricsDataMetricsType
    attributes: List[KeyValue]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        timestamp = self.timestamp
        count = self.count
        metrics_type = self.metrics_type.value

        attributes = []
        for attributes_item_data in self.attributes:
            attributes_item = attributes_item_data.to_dict()

            attributes.append(attributes_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "timestamp": timestamp,
                "count": count,
                "metricsType": metrics_type,
                "attributes": attributes,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        timestamp = d.pop("timestamp")

        count = d.pop("count")

        metrics_type = MetricsDataMetricsType(d.pop("metricsType"))

        attributes = []
        _attributes = d.pop("attributes")
        for attributes_item_data in _attributes:
            attributes_item = KeyValue.from_dict(attributes_item_data)

            attributes.append(attributes_item)

        metrics_data = cls(
            timestamp=timestamp,
            count=count,
            metrics_type=metrics_type,
            attributes=attributes,
        )

        metrics_data.additional_properties = d
        return metrics_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Client for interacting with Harness FF server"""

import threading
from typing import Any, Callable, Dict, Optional

from jwt import decode

from featureflags.analytics import AnalyticsService
from featureflags.evaluations.evaluator import Evaluator
from featureflags.repository import Repository

from .api.client import AuthenticatedClient, Client
from .api.default.authenticate import AuthenticationRequest
from .api.default.authenticate import sync as authenticate
from .config import Config, default_config
from .evaluations.auth_target import Target
from .polling import PollingProcessor
from .streaming import StreamProcessor
from .util import log

VERSION: str = "1.0"


class CfClient(object):
    def __init__(
            self, sdk_key: str,
            *options: Callable,
            config: Optional[Config] = None
    ):
        self._client: Optional[Client] = None
        self._auth_token: Optional[str] = None
        self._environment_id: Optional[str] = None
        self._sdk_key: Optional[str] = sdk_key
        self._config: Config = default_config
        self._cluster: str = '1'

        if config:
            self._config = config

        for option in options:
            if callable(option):
                option(self._config)

        if self._config.cache is None:
            raise Exception("cache cannot be none")

        self._repository = Repository(self._config.cache)
        self._evaluator = Evaluator(self._repository)

        log.debug("CfClient initialized")
        self.run()

    def run(self):
        self.authenticate()

        streaming_event = threading.Event()
        polling_event = threading.Event()

        self._polling_processor = PollingProcessor(
            client=self._client,
            config=self._config,
            environment_id=self._environment_id,
            ready=polling_event,
            stream_ready=streaming_event,
            repository=self._repository
        )
        self._polling_processor.start()

        if self._config.enable_stream:
            self._stream = StreamProcessor(
                repository=self._repository,
                client=self._client,
                environment_id=self._environment_id,
                api_key=self._sdk_key,
                token=self._auth_token,
                config=self._config,
                ready=streaming_event,
                cluster=self._cluster,
            )
            self._stream.start()

        if self._config.enable_analytics:
            self._analytics = AnalyticsService(
                config=self._config,
                client=self._client,
                environment=self._environment_id
            )

    def get_environment_id(self):
        return self._environment_id

    def authenticate(self):
        client = Client(
            base_url=self._config.base_url,
            events_url=self._config.events_url
        )
        body = AuthenticationRequest(api_key=self._sdk_key)
        response = authenticate(client=client, json_body=body)
        self._auth_token = response.auth_token

        decoded = decode(self._auth_token, options={
            "verify_signature": False})
        self._environment_id = decoded["environment"]
        self._cluster = decoded["clusterIdentifier"]
        if not self._cluster:
            self._cluster = '1'
        self._client = AuthenticatedClient(
            base_url=self._config.base_url,
            events_url=self._config.events_url,
            token=self._auth_token,
            params={
                'cluster': self._cluster
            }
        )
        self._client.with_headers({"User-Agent": "PythonSDK/" + VERSION})

    def bool_variation(self, identifier: str, target: Target,
                       default: bool) -> bool:
        variation = self._evaluator.evaluate(identifier, target)
        if self._config.enable_analytics:
            self._analytics.enqueue(target, identifier, variation)
        return variation.bool(default)

    def int_variation(self, identifier: str, target: Target,
                      default: int) -> int:
        variation = self._evaluator.evaluate(identifier, target)
        if self._config.enable_analytics:
            self._analytics.enqueue(target, identifier, variation)
        return variation.int(default)

    def number_variation(self, identifier: str, target: Target,
                         default: float) -> float:
        variation = self._evaluator.evaluate(
            identifier, target)
        if self._config.enable_analytics:
            self._analytics.enqueue(target, identifier, variation)
        return variation.number(default)

    def string_variation(self, identifier: str, target: Target,
                         default: str) -> str:
        variation = self._evaluator.evaluate(
            identifier, target)
        if self._config.enable_analytics:
            self._analytics.enqueue(target, identifier, variation)
        return variation.string(default)

    def json_variation(self, identifier: str, target: Target,
                       default: Dict[str, Any]) -> Dict[str, Any]:
        variation = self._evaluator.evaluate(identifier, target)
        if self._config.enable_analytics:
            self._analytics.enqueue(target, identifier, variation)
        return variation.json(default)

    def close(self):
        log.info('closing sdk client')
        self._polling_processor.stop()
        if self._config.enable_stream:
            self._stream.stop()

        if self._config.enable_analytics:
            self._analytics.close()
        log.info('All threads finished')

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



import pytest

from featureflags.evaluations.auth_target import Target
from featureflags.evaluations.clause import Clause
from featureflags.evaluations.constants import EQUAL_OPERATOR
from featureflags.evaluations.distribution import Distribution
from featureflags.evaluations.enum import FeatureState
from featureflags.evaluations.evaluator import Evaluator
from featureflags.evaluations.feature import FeatureConfig, FeatureConfigKind
from featureflags.evaluations.segment import Segment
from featureflags.evaluations.serve import Serve
from featureflags.evaluations.serving_rule import ServingRule
from featureflags.evaluations.target_map import TargetMap
from featureflags.evaluations.variation import Variation
from featureflags.evaluations.variation_map import VariationMap
from featureflags.evaluations.weighted_variation import WeightedVariation
from featureflags.lru_cache import LRUCache
from featureflags.repository import Repository

TRUE = "true"
FALSE = "false"


@pytest.fixture
def target():
    return Target(identifier="john", name="John",
                  attributes={"email": "john@doe.com"})


@pytest.fixture
def true_variation():
    return Variation(identifier=TRUE, value=TRUE)


@pytest.fixture
def false_variation():
    return Variation(identifier=FALSE, value=FALSE)


@pytest.fixture
def empty_variation():
    return Variation(identifier="", value=None)


@pytest.fixture
def variations(true_variation, false_variation):
    return [true_variation, false_variation]


@pytest.fixture
def distribution_by_email():
    variations = [
        WeightedVariation(TRUE, 50),
        WeightedVariation(FALSE, 50)
    ]
    return Distribution("email", variations)


@pytest.fixture
def feature(variations):

    default_serve = Serve(variation=TRUE)

    return FeatureConfig(
        feature="bool-flag",
        environment="test",
        default_serve=default_serve,
        kind=FeatureConfigKind.BOOLEAN,
        off_variation=FALSE,
        project="default",
        state=FeatureState.ON,
        variations=variations
    )


@pytest.fixture
def segment(target):

    return Segment(
        identifier="beta",
        name="Beta users",
        environment="test",
        included=[Target(identifier=target.identifier, name=target.name)],
        version=1
    )


@pytest.fixture
def data_provider(feature, segment):
    cache = LRUCache()

    repository = Repository(cache)
    repository.set_flag(feature)
    repository.set_segment(segment)

    return repository


@pytest.mark.parametrize('identifier,expected', [
    (TRUE, 'true_variation'),
    (FALSE, 'false_variation'),
    ("unknown", 'empty_variation')
])
def test_find_variation(request, data_provider, variations, identifier,
                        expected):
    evaluator = Evaluator(data_provider)

    expected = request.getfixturevalue(expected)

    got = evaluator._find_variation(variations, identifier)

    assert got == expected


@pytest.mark.parametrize('bucket_by,value,expected', [
    ("email", 'john@doe.com', 18),
    ("email", 'enver.bisevac@harness', 61),
    ("identifier", 'harness', 10)
])
def test_get_normalized_number(bucket_by, value, expected):
    evaluator = Evaluator(data_provider)

    got = evaluator._get_normalized_number(bucket_by, value)

    assert got == expected


def test_is_enabled(data_provider, target):
    evaluator = Evaluator(data_provider)

    got = evaluator._is_enabled(target, "email", 10)

    assert got is False


def test_evaluate_distribution(data_provider, distribution_by_email, target):
    evaluator = Evaluator(data_provider)

    got = evaluator._evaluate_distribution(distribution_by_email, target)

    assert got == 'true'


def test_check_target_in_segment(data_provider, target):
    evaluator = Evaluator(data_provider)

    got = evaluator._check_target_in_segment(["beta"], target)

    assert got is True


def test_evaluate_clause(data_provider, target):
    evaluator = Evaluator(data_provider)

    clause = Clause(
        id="",
        attribute="identifier",
        op=EQUAL_OPERATOR,
        values=[target.identifier]
    )

    got = evaluator._evaluate_clause(clause, target)

    assert got is True


def test_evaluate_rules(data_provider, target):
    evaluator = Evaluator(data_provider)

    clause = Clause(
        id="",
        attribute="identifier",
        op=EQUAL_OPERATOR,
        values=[target.identifier]
    )

    rules = [
        ServingRule(
            clauses=[clause],
            priority=1000,
            rule_id="",
            serve=Serve(
                variation=TRUE
            )
        )
    ]

    got = evaluator._evaluate_rules(rules, target)

    assert got == TRUE


def test_evaluate_variation_map(data_provider, target):
    evaluator = Evaluator(data_provider)

    vmap = [
        VariationMap(variation=TRUE, targets=[TargetMap(
            name=target.name, identifier=target.identifier)])
    ]

    got = evaluator._evaluate_variation_map(vmap, target)

    assert got == TRUE


def test_evaluate_variation_segments_map(data_provider, target):
    evaluator = Evaluator(data_provider)

    vmap = [
        VariationMap(variation=TRUE, target_segments=["beta"])
    ]

    got = evaluator._evaluate_variation_map(vmap, target)

    assert got == TRUE


def test_evaluate_flag_off(data_provider, feature, target, false_variation):
    evaluator = Evaluator(data_provider)
    feature.state = FeatureState.OFF
    got = evaluator._evaluate_flag(feature, target)

    assert got == false_variation


def test_evaluate_flag_on(data_provider, feature, target, true_variation):
    evaluator = Evaluator(data_provider)

    got = evaluator._evaluate_flag(feature, target)

    assert got == true_variation



"""Type utilities"""

import typing


def get_value(value: typing.Union[typing.Any,
              typing.List[typing.Any]]) -> typing.Any:
    if isinstance(value, list):
        return value[0]
    return value


def get_str_value(value: typing.List[typing.Any]) -> str:
    return str(get_value(value))


def get_int_value(value: typing.List[typing.Any]) -> int:
    return int(get_value(value))


def get_float_value(value: typing.List[typing.Any]) -> float:
    return float(get_value(value))



#!/usr/bin/env python

"""The setup script."""

from sys import version
from setuptools import find_packages, setup


with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "httpx~=0.16.1",
    "pyjwt~=2.0.1",
    "attrs~=21.2.0",
    "mmh3~=3.0.0",
    "requests==2.25.1"
]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Enver Bisevac",
    author_email="enver.bisevac@harness.io",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Feature flag server SDK for python",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n\n" + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords="featureflags",
    name="harness-featureflags",
    packages=find_packages(include=["featureflags", "featureflags.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/drone/ff-python-server-sdk",
    version='1.1.1',
    zip_safe=False,
)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="Pagination")


@attr.s(auto_attribs=True)
class Pagination:
    """ """

    page_count: int
    item_count: int
    page_size: int
    page_index: int
    version: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        page_count = self.page_count
        item_count = self.item_count
        page_size = self.page_size
        page_index = self.page_index
        version = self.version

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "pageCount": page_count,
                "itemCount": item_count,
                "pageSize": page_size,
                "pageIndex": page_index,
            }
        )
        if version is not UNSET:
            field_dict["version"] = version

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        page_count = d.pop("pageCount")

        item_count = d.pop("itemCount")

        page_size = d.pop("pageSize")

        page_index = d.pop("pageIndex")

        version = d.pop("version", UNSET)

        pagination = cls(
            page_count=page_count,
            item_count=item_count,
            page_size=page_size,
            page_index=page_index,
            version=version,
        )

        pagination.additional_properties = d
        return pagination

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections import namedtuple

import pytest

from featureflags.lru_cache import LRUCache


def test_cache_null():
    """Null dict is null."""
    cache = LRUCache()
    assert cache.__len__() == 0


Case = namedtuple("Case", ["size", "len", "init"])


@pytest.mark.parametrize(
    "case",
    [
        Case(9, 0, []),
        Case(9, 1, [("one", 1)]),
        Case(9, 2, [("one", 1), ("two", 2)]),
        Case(2, 2, [("one", 1), ("two", 2)]),
        Case(1, 1, [("one", 1), ("two", 2)]),
    ],
)
@pytest.mark.parametrize("method", ["assign", "init"])
def test_cache_init(case, method):
    """Check that the # of elements is right, given # given and size."""
    if method == "init":
        cache = LRUCache(case.init, size=case.size)
    elif method == "assign":
        cache = LRUCache(size=case.size)
        for (key, val) in case.init:
            cache[key] = val
    else:
        assert False

    # length is max(#entries, size)
    assert cache.__len__() == case.len

    # make sure the first entry is the one ejected
    if case.size > 1 and case.init:
        assert "one" in cache.keys()
    else:
        assert "one" not in cache.keys()


@pytest.mark.parametrize("method", ["init", "assign"])
def test_cache_overflow_default(method):
    """Test default overflow logic."""
    if method == "init":
        cache = LRUCache([("one", 1), ("two", 2), ("three", 3)], size=2)
    elif method == "assign":
        cache = LRUCache(size=2)
        cache["one"] = 1
        cache["two"] = 2
        cache["three"] = 3
    else:
        assert False

    assert "one" not in cache.keys()
    assert "two" in cache.keys()
    assert "three" in cache.keys()


@pytest.mark.parametrize("mode", ["get", "set"])
@pytest.mark.parametrize("add_third", [True, False])
def test_cache_lru_overflow(mode, add_third):
    """Test that key access resets LRU logic."""

    cache = LRUCache([("one", 1), ("two", 2)], size=2)

    if mode == "get":
        _ = cache["one"]
    elif mode == "set":
        cache["one"] = 1
    else:
        assert False

    if add_third:
        cache["three"] = 3

        assert "one" in cache.keys()
        assert "two" not in cache.keys()
        assert "three" in cache.keys()
    else:
        assert "one" in cache.keys()
        assert "two" in cache.keys()
        assert "three" not in cache.keys()


def test_cache_keyerror():
    cache = LRUCache()
    with pytest.raises(KeyError):
        cache["foo"]


def test_cache_miss_doesnt_eject():
    cache = LRUCache([("one", 1), ("two", 2)], size=2)
    with pytest.raises(KeyError):
        cache["foo"]

    assert len(cache) == 2
    assert "one" in cache.keys()
    assert "two" in cache.keys()



from typing import Any, Dict, List, Type, TypeVar

import attr

from .auth_target import Target
from .constants import ONE_HUNDRED
from .strategy import get_normalized_number
from .weighted_variation import WeightedVariation

T = TypeVar("T", bound="Distribution")


@attr.s(auto_attribs=True)
class Distribution(object):
    bucket_by: str
    variations: List[WeightedVariation]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def get_key_name(self, target: Target) -> str:
        variation: str = ""
        for _variation in self.variations:
            variation = _variation.variation
            if self.is_enabled(target, _variation.weight):
                return variation
        if self.is_enabled(target, ONE_HUNDRED):
            return variation
        return ""

    def is_enabled(self, target: Target, percentage: int) -> bool:
        identifier = target.get_attr_value(self.bucket_by)
        if not identifier or identifier == "":
            return False
        bucket_id = get_normalized_number(identifier, self.bucket_by)
        return percentage > 0 and bucket_id <= percentage

    def to_dict(self) -> Dict[str, Any]:
        bucket_by = self.bucket_by
        variations = []
        for variations_item_data in self.variations:
            variations_item = variations_item_data.to_dict()

            variations.append(variations_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "bucketBy": bucket_by,
                "variations": variations,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        if src_dict is None:
            return None
        d = src_dict.copy()
        bucket_by = d.pop("bucketBy")

        variations = []
        _variations = d.pop("variations")
        for variations_item_data in _variations:
            variations_item = WeightedVariation.from_dict(variations_item_data)

            variations.append(variations_item)

        distribution = cls(
            bucket_by=bucket_by,
            variations=variations,
        )

        distribution.additional_properties = d
        return distribution

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
